# frozen_string_literal: true

module Dry
  module Types
    # Internal container for the built-in types
    #
    # @api private
    class Container
      include Core::Container::Mixin
    end
  end
end
